@@ -1,15 +1,17 @@
-from xarray import Dataset, DataArray, concat
-from xarray.core.coordinates import DatasetCoordinates
-from roguewave.tools.math import wrapped_difference
 import numpy
-from typing import Iterator, Hashable, Iterable
-from roguewave.interpolate.dataset import interpolate_dataset_grid, \
+from roguewave.interpolate.dataset import (
+    interpolate_dataset_grid,
     interpolate_dataset_along_axis
+)
+from roguewave.tools.math import wrapped_difference
+from roguewave.tools.time import to_datetime64
 from roguewave.wavetheory.lineardispersion import \
     inverse_intrinsic_dispersion_relation
-from roguewave.tools.time import to_datetime64
 from roguewave.wavespectra.estimators import mem2
 from roguewave.wavespectra.parametric import pierson_moskowitz_frequency
+from typing import Iterator, Hashable, Iterable, TypeVar
+from xarray import Dataset, DataArray, concat
+from xarray.core.coordinates import DatasetCoordinates
 
 _NAME_F = 'frequency'
 _NAME_D = 'direction'
@@ -26,8 +28,17 @@ _NAMES_2D = (_NAME_F, _NAME_D, _NAME_T, _NAME_E, _NAME_LAT, _NAME_LON)
 _NAMES_1D = (_NAME_F, _NAME_T, _NAME_E, _NAME_LAT, _NAME_LON,
              _NAME_a1, _NAME_b1, _NAME_a2, _NAME_b2)
 
+_T = TypeVar("_T")
 
 class DatasetWrapper():
+    """
+    A class that wraps a dataset object and passes through some of its primary
+    functionality (get/set etc.). Used here mostly to make explicit what parts
+    of the Dataset interface we actually expose in frequency objects. Note that
+    we do not claim- or try to obtain completeness here. If full capabilities
+    of the dataset object are needed we can simple operate directly on the
+    dataset object itself.
+    """
     def __init__(self, dataset: Dataset):
         self.dataset = dataset
 
@@ -37,17 +48,20 @@ class DatasetWrapper():
     def __setitem__(self, key, value) -> None:
         return self.dataset.__setitem__(key, value)
 
-    def __copy__(self):
+    def __copy__(self:_T) -> _T:
         cls = self.__class__
         return cls(self.dataset.copy())
 
-    def copy(self, deep=True):
+    def __len__(self):
+        return len(self.dataset)
+
+    def copy(self:_T, deep=True) -> _T:
         if deep:
             return self.__deepcopy__({})
         else:
             return self.__copy__()
 
-    def __deepcopy__(self, memodict):
+    def __deepcopy__(self:_T, memodict) -> _T:
         cls = self.__class__
         return cls(self.dataset.copy(deep=True))
 
@@ -63,12 +77,28 @@ class DatasetWrapper():
     def __iter__(self) -> Iterator[Hashable]:
         return self.dataset.__iter__()
 
+    @classmethod
+    def concat_from_list(cls: _T, _list: Iterable[_T], dim='time') -> _T:
+        """
+        Classmethod to aggregate a list of wavespectra and concatenate them
+        into a single spectral object. For instance, from spotter observations
+        we may get 1 spectrum per timestamp- and we want to aggregate all
+        spectra in a single object.
+
+        :param _list: list (Iterable) of WaveSpectra
+        :param dim: the dimension over which to concatenate the spectra,
+            usually time.
+        :return: Wavespectrum object.
+        """
+        return cls(concat( [x.dataset for x in _list], dim=dim))
+
 
 class WaveSpectrum(DatasetWrapper):
     frequency_units = 'Hertz'
     angular_units = 'Degrees'
     spectral_density_units = 'm**2/Hertz'
-    angular_convention = 'Wave travel direction (going-to), measured anti-clockwise from East'
+    angular_convention = 'Wave travel direction (going-to), ' \
+                         'measured anti-clockwise from East'
     bulk_properties = (
         'm0', 'hm0', 'tm01', 'tm02', 'peak_period', 'peak_direction',
         'peak_directional_spread', 'mean_direction', 'mean_directional_spread',
@@ -79,109 +109,254 @@ class WaveSpectrum(DatasetWrapper):
     def __init__(self, dataset: Dataset):
         super(WaveSpectrum, self).__init__(dataset)
 
+    def __add__(self:_T, other: _T) -> _T:
+        spectrum = self.copy(deep=True)
+        spectrum[_NAME_E] = (spectrum[_NAME_E] +
+                             other[_NAME_E])
+        return spectrum
+
+    def __sub__(self: _T, other: _T) -> _T:
+        spectrum = self.copy(deep=True)
+        spectrum[_NAME_E] = (spectrum[_NAME_E] -
+                             other[_NAME_E])
+        return spectrum
+
+    def __neg__(self:_T) -> _T:
+        """
+        Negate self- that is -spectrum
+        :return: spectrum with all spectral values taken to have the opposite
+            sign.
+        """
+        spectrum = self.copy(deep=True)
+        spectrum[_NAME_E] = -spectrum[_NAME_E]
+        return spectrum
+
     def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf
                          ) -> DataArray:
-        _range = self._range(fmin, fmax)
-        return (self.e.isel({_NAME_F: _range}) * self.frequency[
-            _range] ** power
-                ).integrate(coord=_NAME_F)
+        """
+        Calculate a "frequency moment" over the given range. A frequency moment
+        here refers to the integral:
 
-    def _range(self, fmin=0.0, fmax=numpy.inf) -> numpy.ndarray:
-        return ((self[_NAME_F].values >= fmin) &
-                (self[_NAME_F].values < fmax))
+                    Integral-over-frequency-range[ e(f) * f**power ]
+
+        :param power: power of the frequency
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: frequency moment
+        """
+        _range = self._range(fmin, fmax)
+        return (
+            self.e.isel({_NAME_F: _range}) * self.frequency[_range] ** power
+        ).integrate(coord=_NAME_F)
 
     @property
-    def number_of_frequencies(self):
+    def number_of_frequencies(self) -> int:
+        """
+        :return: number of frequencies
+        """
         return len(self.frequency)
 
     @property
-    def spectral_values(self):
+    def spectral_values(self) -> DataArray:
+        """
+        :return: Spectral levels
+        """
         return self.dataset[_NAME_E]
 
     @property
     def radian_frequency(self) -> DataArray:
+        """
+        :return: Radian frequency
+        """
         data_array = self[_NAME_F] * 2 * numpy.pi
         data_array.name = 'radian_frequency'
         return data_array
 
     @property
     def latitude(self) -> DataArray:
+        """
+        :return: latitudes
+        """
         return self[_NAME_LAT]
 
     @property
     def longitude(self) -> DataArray:
+        """
+        :return: longitudes
+        """
         return self[_NAME_LON]
 
     @property
     def time(self) -> DataArray:
+        """
+        :return: Time
+        """
         return self[_NAME_T]
 
     @property
     def e(self) -> DataArray:
+        """
+        :return: 1D spectral values (directionally integrated spectrum).
+            Equivalent to self.spectral_values if this is a 1D spectrum.
+        """
         return self[_NAME_E]
 
     @property
     def a1(self) -> DataArray:
+        """
+        :return: normalized Fourier moment cos(theta)
+        """
         return self[_NAME_a1]
 
     @property
     def b1(self) -> DataArray:
+        """
+        :return: normalized Fourier moment sin(theta)
+        """
         return self[_NAME_b1]
 
     @property
     def a2(self) -> DataArray:
+        """
+        :return: normalized Fourier moment cos(2*theta)
+        """
         return self[_NAME_a2]
 
     @property
     def b2(self) -> DataArray:
+        """
+        :return: normalized Fourier moment sin(2*theta)
+        """
         return self[_NAME_b2]
 
     @property
     def A1(self) -> DataArray:
+        """
+        :return: Fourier moment cos(theta)
+        """
         return self.a1 * self.e
 
     @property
     def B1(self) -> DataArray:
+        """
+        :return: Fourier moment sin(theta)
+        """
         return self.b1 * self.e
 
     @property
     def A2(self) -> DataArray:
+        """
+        :return: Fourier moment cos(2*theta)
+        """
         return self.a2 * self.e
 
     @property
     def B2(self) -> DataArray:
+        """
+        :return: Fourier moment sin(2*theta)
+        """
         return self.b2 * self.e
 
     @property
     def frequency(self) -> DataArray:
+        """
+        :return: Frequencies (Hz)
+        """
         return self[_NAME_F]
 
     def m0(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        Zero order frequency moment. Also referred to as variance or energy.
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: variance/energy
+        """
         return self.frequency_moment(0, fmin, fmax)
 
     def m1(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        First order frequency moment. Primarily used in calculating a mean
+        period measure (Tm01)
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: first order frequency moment.
+        """
         return self.frequency_moment(1, fmin, fmax)
 
     def m2(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        Second order frequency moment. Primarily used in calculating the zero
+        crossing period (Tm02)
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: Second order frequency moment.
+        """
         return self.frequency_moment(2, fmin, fmax)
 
     def hm0(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        Significant wave height estimated from the spectrum, i.e. waveheight
+        h estimated from variance m0. Common notation in literature.
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: Significant wave height
+        """
         return 4 * numpy.sqrt(self.m0(fmin, fmax))
 
     def tm01(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        Mean period, estimated as the inverse of the center of mass of the
+        spectral curve under the 1d spectrum.
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: Mean period
+        """
         return self.m0(fmin, fmax) / self.m1(fmin, fmax)
 
     def tm02(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        Zero crossing period based on Rice's spectral estimate.
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: Zero crossing period
+        """
         return numpy.sqrt(self.m0(fmin, fmax) / self.m2(fmin, fmax))
 
     def peak_index(self, fmin=0, fmax=numpy.inf) -> DataArray:
-        range = self._range(fmin, fmax)
-        return numpy.argmax(self.e.where(range, 0).argmax(dim=_NAME_F))
+        """
+        Index of the peak frequency of the 1d spectrum within the given range
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: peak indices
+        """
+        return self.e.where(self._range(fmin, fmax), 0).argmax(dim=_NAME_F)
 
     def peak_frequency(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        Peak frequency of the spectrum, i.e. frequency at which the spectrum
+        obtains its maximum.
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: peak frequency
+        """
         return self[_NAME_F][self.peak_index(fmin, fmax)]
 
     def peak_period(self, fmin=0, fmax=numpy.inf) -> DataArray:
+        """
+        Peak period of the spectrum, i.e. period at which the spectrum
+        obtains its maximum.
+
+        :param fmin: minimum frequency
+        :param fmax: maximum frequency
+        :return: peak period
+        """
         return 1 / self.peak_frequency(fmin, fmax)
 
     def peak_direction(self, fmin=0, fmax=numpy.inf) -> DataArray:
@@ -190,27 +365,27 @@ class WaveSpectrum(DatasetWrapper):
             self.a1.isel(**{_NAME_F: index}),
             self.a1.isel(**{_NAME_F: index}))
 
-    def peak_directional_spread(self, fmin=0, fmax=numpy.inf):
+    def peak_directional_spread(self, fmin=0, fmax=numpy.inf) -> DataArray:
         index = self.peak_index(fmin, fmax)
         a1 = self.a1.isel(**{_NAME_F: index})
         b1 = self.a1.isel(**{_NAME_F: index})
         return self._spread(a1, b1)
 
     @staticmethod
-    def _mean_direction(a1, b1):
+    def _mean_direction(a1:_T, b1:_T) -> _T:
         return numpy.arctan2(b1, a1) * 180 / numpy.pi
 
     @staticmethod
-    def _spread(a1, b1):
+    def _spread(a1:_T, b1:_T) -> _T:
         return numpy.sqrt(
             2 - 2 * numpy.sqrt(a1 ** 2 + b1 ** 2)) * 180 / numpy.pi
 
     @property
-    def mean_direction_per_frequency(self):
+    def mean_direction_per_frequency(self) -> DataArray:
         return self._mean_direction(self.a1, self.b1)
 
     @property
-    def mean_spread_per_frequency(self):
+    def mean_spread_per_frequency(self) -> DataArray:
         return self._spread(self.a1, self.b1)
 
     def _spectral_weighted(self, property: DataArray, fmin=0, fmax=numpy.inf):
@@ -245,11 +420,17 @@ class WaveSpectrum(DatasetWrapper):
         else:
             return numpy.inf
 
-    def wavenumber(self, **kwargs):
+    @property
+    def wavenumber(self) -> DataArray:
         return inverse_intrinsic_dispersion_relation(self.radian_frequency,
-                                                     self.depth, **kwargs)
+                                                     self.depth)
 
-    def peak_wavenumber(self):
+    @property
+    def wavelength(self) -> DataArray:
+        return 2 * numpy.pi / self.wavenumber
+
+    @property
+    def peak_wavenumber(self) -> DataArray:
         index = self.peak_index()
         return inverse_intrinsic_dispersion_relation(
             self.radian_frequency[index], self.depth)
@@ -270,43 +451,24 @@ class WaveSpectrum(DatasetWrapper):
         return dataset
 
     @property
-    def significant_waveheight(self):
+    def significant_waveheight(self) -> DataArray:
         return self.hm0()
 
     @property
-    def mean_period(self):
+    def mean_period(self) -> DataArray:
         return self.tm01()
 
-    def __len__(self):
-        return len(self.dataset)
+    @property
+    def zero_crossing_period(self) -> DataArray:
+        return self.tm02()
 
-    def __add__(self, other: "WaveSpectrum") -> "WaveSpectrum":
-        spectrum = self.copy(deep=True)
-        spectrum[_NAME_E] = (spectrum[_NAME_E] +
-                             other[_NAME_E])
-        return spectrum
+    def interpolate(self:_T, coordinates) -> _T:
+        return self.__class__(
+            interpolate_dataset_grid(coordinates, self.dataset))
 
-    def __sub__(self, other: "WaveSpectrum") -> "WaveSpectrum":
-        spectrum = self.copy(deep=True)
-        spectrum[_NAME_E] = (spectrum[_NAME_E] -
-                             other[_NAME_E])
-        return spectrum
-
-    def __neg__(self, other: "WaveSpectrum") -> "WaveSpectrum":
-        spectrum = self.copy(deep=True)
-        spectrum[_NAME_E] = -spectrum[_NAME_E]
-        return spectrum
-
-    def interpolate(self, coordinates):
-        self.dataset = interpolate_dataset_grid(
-            coordinates, self.dataset
-        )
-        return self
-
-    def interpolate_frequency(self, new_frequencies):
-        self.dataset = interpolate_dataset_along_axis(
-            new_frequencies, self.dataset, coordinate_name='frequency')
-        return self
+    def interpolate_frequency(self:_T, new_frequencies) -> _T:
+        return self.__class__(interpolate_dataset_along_axis(
+            new_frequencies, self.dataset, coordinate_name='frequency'))
 
     @property
     def is_sea_spectrum(self) -> DataArray:
@@ -333,12 +495,9 @@ class WaveSpectrum(DatasetWrapper):
     def is_swell_spectrum(self) -> DataArray:
         return ~self.is_sea_spectrum
 
-    @classmethod
-    def concat_from_list(cls,
-                         _list: Iterable["WaveSpectrum"],
-                         dim='time') -> "WaveSpectrum":
-        _data = [x.dataset for x in _list]
-        return cls(concat(_data, dim=dim))
+    def _range(self, fmin=0.0, fmax=numpy.inf) -> numpy.ndarray:
+        return ((self[_NAME_F].values >= fmin) &
+                (self[_NAME_F].values < fmax))
 
 
 class FrequencyDirectionSpectrum(WaveSpectrum):
@@ -397,7 +556,7 @@ class FrequencyDirectionSpectrum(WaveSpectrum):
     @property
     def b2(self) -> DataArray:
         return self._directionally_integrate(
-            self[_NAME_E] * numpy.cos(
+            self[_NAME_E] * numpy.sin(
                 2 * self.radian_direction)
         )
 
@@ -419,7 +578,7 @@ class FrequencyDirectionSpectrum(WaveSpectrum):
         )
 
     @property
-    def number_of_directions(self):
+    def number_of_directions(self) -> int:
         return len(self.direction)
 
 
@@ -504,18 +663,19 @@ def create_1d_spectrum(
         a2 = a2[None, :]
         b2 = b2[None, :]
 
-    return FrequencySpectrum(Dataset(
-        data_vars={
-            _NAME_E: ((_NAME_T, _NAME_F), variance_density),
-            _NAME_a1: ((_NAME_T, _NAME_F), a1),
-            _NAME_b1: ((_NAME_T, _NAME_F), b1),
-            _NAME_a2: ((_NAME_T, _NAME_F), a2),
-            _NAME_b2: ((_NAME_T, _NAME_F), b2),
-            _NAME_LAT: ((_NAME_T,), latitude),
-            _NAME_LON: ((_NAME_T,), longitude),
-            _NAME_DEPTH: ((_NAME_T,), depth)
-        },
-        coords={_NAME_T: to_datetime64(time), _NAME_F: frequency}
+    return FrequencySpectrum(
+        Dataset(
+            data_vars={
+                _NAME_E: ((_NAME_T, _NAME_F), variance_density),
+                _NAME_a1: ((_NAME_T, _NAME_F), a1),
+                _NAME_b1: ((_NAME_T, _NAME_F), b1),
+                _NAME_a2: ((_NAME_T, _NAME_F), a2),
+                _NAME_b2: ((_NAME_T, _NAME_F), b2),
+                _NAME_LAT: ((_NAME_T,), latitude),
+                _NAME_LON: ((_NAME_T,), longitude),
+                _NAME_DEPTH: ((_NAME_T,), depth)
+            },
+            coords={_NAME_T: to_datetime64(time), _NAME_F: frequency}
         )
     )
 
