@@ -336,7 +336,8 @@ class RestartFile(Sequence):
                 number_of_frequencies,
                 number_of_directions)
         """
-
+        latitude = numpy.atleast_1d(latitude)
+        longitude = numpy.atleast_1d(longitude)
         points = {"latitude":numpy.atleast_1d(latitude),
                   "longitude":numpy.atleast_1d(longitude)}
 
@@ -370,7 +371,7 @@ class RestartFile(Sequence):
             data_coordinates=data_coordinates,
             data_shape=data_shape,
             interp_coord_names=list(points.keys()),
-            interp_index_coord_name='time',
+            interp_index_coord_name='latitude',
             data_periodic_coordinates=periodic_coordinates,
             data_period=None,
             data_discont=None
@@ -392,7 +393,7 @@ class RestartFile(Sequence):
         To note, we add the trailing 1 to make it easy to use broadcasting (
             the Jacobian is constant in directional space).
         """
-        return 1 / self.to_frequency_energy_density(s)[:,:,None]
+        return 1 / self.to_frequency_energy_density(s)
 
     def to_frequency_energy_density(
             self, s:Union[slice, numpy.ndarray, Sequence]
@@ -512,7 +513,7 @@ class RestartFile(Sequence):
         toggle = not self._convert
         if toggle:
             self.set_return_freq_energy_density()
-        spectra = self['variance_density'][index].values
+        spectra = self[index]['variance_density'].values
         if toggle:
             self.set_return_k_action_density()
 
@@ -721,7 +722,7 @@ class RestartFileStack:
             )
         self._update_progress_bar(len(_input))
 
-        data = concat( data , dim='time',  )
+        data = concat( data , dim='time')
         data.coords['time'] = to_datetime64([self.time[it] for it in time_index])
         return FrequencyDirectionSpectrum(data)
 
